{$ALIGN ON, $BOOLEVAL OFF, $LONGSTRINGS ON, $IOCHECKS ON, $WRITEABLECONST OFF}
{$OVERFLOWCHECKS OFF, $RANGECHECKS OFF, $TYPEDADDRESS ON, $MINENUMSIZE 1}
unit MHTools;

{ MHTools Version 1.02

  Copyright:
     2000, Peter J. Haas, Barleben, Germany

    Bitte License.txt lesen;  Please read License.txt

  Kontakt/Contact:
    EMail:     PeterJHaas@t-online.de,
    HomePage:  http://home.t-online.de/home/PeterJHaas/delphi.htm

  Beschreibung:
    Die Tools dienen der Umsetzung von MH (Modified Huffman) fr das
    Structured Fax File - Format

  Description:
    These Tools are for the permutation of MH (Modified Huffman) for the
    Structured Fax File format

  History:
    Bitte Readme.txt lesen;  Please read Readme.txt

  Quellen/Source:
    RFC 804  CCITT Draft Recommendation T.4
    TIFF-Spezification, Revision 6.0, Final  June 3, 1992
}

interface

// Termination-Code ermitteln;  Detect termination code
function MHSearchTerminationCode(ABits:Word; var ABitCount, APixelCount:Integer;
                                 ABlackPixels:Boolean):Boolean;
// MakeUp-Code ermitteln;  Detect MakeUp code
function MHSearchMakeUp(ABits:Word; var ABitCount, APixelCount:Integer;
                        ABlackPixels:Boolean):Boolean;

implementation

type
  TSFFCodes = packed record
    Code : Word;
    Lng  : Word;
  end;

  PSFFMHTC = ^TSFFMHTC;
  TSFFMHTC = array[0..63] of TSFFCodes;

  PSFFMHMU = ^TSFFMHMU;
  TSFFMHMU = array[0..26] of TSFFCodes;

  TSFFMHEMU = array[0..12] of TSFFCodes;

const
  // wei Terminierungscode;  white termination code
  cSFFMHWTC : TSFFMHTC = (
    (Code: $AC; Lng: 8),   // 0011 0101        8       0
    (Code: $38; Lng: 6),   // 0001 11          6       1
    (Code: $0E; Lng: 4),   // 0111             4       2
    (Code: $01; Lng: 4),   // 1000             4       3
    (Code: $0D; Lng: 4),   // 1011             4       4
    (Code: $03; Lng: 4),   // 1100             4       5
    (Code: $07; Lng: 4),   // 1110             4       6
    (Code: $0F; Lng: 4),   // 1111             4       7
    (Code: $19; Lng: 5),   // 1001 1           5       8
    (Code: $05; Lng: 5),   // 1010 0           5       9
    (Code: $1C; Lng: 5),   // 0011 1           5       10
    (Code: $02; Lng: 5),   // 0100 0           5       11
    (Code: $04; Lng: 6),   // 0010 00          6       12
    (Code: $30; Lng: 6),   // 0000 11          6       13
    (Code: $0B; Lng: 6),   // 1101 00          6       14
    (Code: $2B; Lng: 6),   // 1101 01          6       15
    (Code: $15; Lng: 6),   // 1010 10          6       16
    (Code: $35; Lng: 6),   // 1010 11          6       17
    (Code: $72; Lng: 7),   // 0100 111         7       18
    (Code: $18; Lng: 7),   // 0001 100         7       19
    (Code: $08; Lng: 7),   // 0001 000         7       20
    (Code: $74; Lng: 7),   // 0010 111         7       21
    (Code: $60; Lng: 7),   // 0000 011         7       22
    (Code: $10; Lng: 7),   // 0000 100         7       23
    (Code: $0A; Lng: 7),   // 0101 000         7       24
    (Code: $6A; Lng: 7),   // 0101 011         7       25
    (Code: $64; Lng: 7),   // 0010 011         7       26
    (Code: $12; Lng: 7),   // 0100 100         7       27
    (Code: $0C; Lng: 7),   // 0011 000         7       28
    (Code: $40; Lng: 8),   // 0000 0010        8       29
    (Code: $C0; Lng: 8),   // 0000 0011        8       30
    (Code: $58; Lng: 8),   // 0001 1010        8       31
    (Code: $D8; Lng: 8),   // 0001 1011        8       32
    (Code: $48; Lng: 8),   // 0001 0010        8       33
    (Code: $C8; Lng: 8),   // 0001 0011        8       34
    (Code: $28; Lng: 8),   // 0001 0100        8       35
    (Code: $A8; Lng: 8),   // 0001 0101        8       36
    (Code: $68; Lng: 8),   // 0001 0110        8       37
    (Code: $E8; Lng: 8),   // 0001 0111        8       38
    (Code: $14; Lng: 8),   // 0010 1000        8       39
    (Code: $94; Lng: 8),   // 0010 1001        8       40
    (Code: $54; Lng: 8),   // 0010 1010        8       41
    (Code: $D4; Lng: 8),   // 0010 1011        8       42
    (Code: $34; Lng: 8),   // 0010 1100        8       43
    (Code: $B4; Lng: 8),   // 0010 1101        8       44
    (Code: $20; Lng: 8),   // 0000 0100        8       45
    (Code: $A0; Lng: 8),   // 0000 0101        8       46
    (Code: $50; Lng: 8),   // 0000 1010        8       47
    (Code: $D0; Lng: 8),   // 0000 1011        8       48
    (Code: $4A; Lng: 8),   // 0101 0010        8       49
    (Code: $CA; Lng: 8),   // 0101 0011        8       50
    (Code: $2A; Lng: 8),   // 0101 0100        8       51
    (Code: $AA; Lng: 8),   // 0101 0101        8       52
    (Code: $24; Lng: 8),   // 0010 0100        8       53
    (Code: $A4; Lng: 8),   // 0010 0101        8       54
    (Code: $1A; Lng: 8),   // 0101 1000        8       55
    (Code: $9A; Lng: 8),   // 0101 1001        8       56
    (Code: $5A; Lng: 8),   // 0101 1010        8       57
    (Code: $DA; Lng: 8),   // 0101 1011        8       58
    (Code: $52; Lng: 8),   // 0100 1010        8       59
    (Code: $D2; Lng: 8),   // 0100 1011        8       60
    (Code: $4C; Lng: 8),   // 0011 0010        8       61
    (Code: $CC; Lng: 8),   // 0011 0011        8       62
    (Code: $2C; Lng: 8));  // 0011 0100        8       63

  // wei MakeUp-Code;  white MakeUp code
  cSFFMHWMU : TSFFMHMU = (
    (Code: $01B; Lng: 5),   // 1101 1            5       64
    (Code: $009; Lng: 5),   // 1001 0            5       128
    (Code: $03A; Lng: 6),   // 0101 11           6       192
    (Code: $076; Lng: 7),   // 0110 111          7       256
    (Code: $06C; Lng: 8),   // 0011 0110         8       320
    (Code: $0EC; Lng: 8),   // 0011 0111         8       384
    (Code: $026; Lng: 8),   // 0110 0100         8       448
    (Code: $0A6; Lng: 8),   // 0110 0101         8       512
    (Code: $016; Lng: 8),   // 0110 1000         8       576
    (Code: $0E6; Lng: 8),   // 0110 0111         8       640
    (Code: $066; Lng: 9),   // 0110 0110 0       9       704
    (Code: $166; Lng: 9),   // 0110 0110 1       9       768
    (Code: $096; Lng: 9),   // 0110 1001 0       9       832
    (Code: $196; Lng: 9),   // 0110 1001 1       9       896
    (Code: $056; Lng: 9),   // 0110 1010 0       9       960
    (Code: $156; Lng: 9),   // 0110 1010 1       9       1024
    (Code: $0D6; Lng: 9),   // 0110 1011 0       9       1088
    (Code: $1D6; Lng: 9),   // 0110 1011 1       9       1152
    (Code: $036; Lng: 9),   // 0110 1100 0       9       1216
    (Code: $136; Lng: 9),   // 0110 1100 1       9       1280
    (Code: $0B6; Lng: 9),   // 0110 1101 0       9       1344
    (Code: $1B6; Lng: 9),   // 0110 1101 1       9       1408
    (Code: $032; Lng: 9),   // 0100 1100 0       9       1472
    (Code: $132; Lng: 9),   // 0100 1100 1       9       1536
    (Code: $0B2; Lng: 9),   // 0100 1101 0       9       1600
    (Code: $006; Lng: 6),   // 0110 00           6       1664
    (Code: $1B2; Lng: 9));  // 0100 1101 1       9       1728


  // schwarz Terminierungscode;  black termination code
  cSFFMHBTC : TSFFMHTC = (
    (Code: $3B0; Lng: 10),   // 0000 1101 11      10      0
    (Code: $002; Lng: 3 ),   // 010               3       1
    (Code: $003; Lng: 2 ),   // 11                2       2
    (Code: $001; Lng: 2 ),   // 10                2       3
    (Code: $006; Lng: 3 ),   // 011               3       4
    (Code: $00C; Lng: 4 ),   // 0011              4       5
    (Code: $004; Lng: 4 ),   // 0010              4       6
    (Code: $018; Lng: 5 ),   // 0001 1            5       7
    (Code: $028; Lng: 6 ),   // 0001 01           6       8
    (Code: $008; Lng: 6 ),   // 0001 00           6       9
    (Code: $010; Lng: 7 ),   // 0000 100          7       10
    (Code: $050; Lng: 7 ),   // 0000 101          7       11
    (Code: $070; Lng: 7 ),   // 0000 111          7       12
    (Code: $020; Lng: 8 ),   // 0000 0100         8       13
    (Code: $0E0; Lng: 8 ),   // 0000 0111         8       14
    (Code: $030; Lng: 9 ),   // 0000 1100 0       9       15
    (Code: $3A0; Lng: 10),   // 0000 0101 11      10      16
    (Code: $060; Lng: 10),   // 0000 0110 00      10      17
    (Code: $040; Lng: 10),   // 0000 0010 00      10      18
    (Code: $730; Lng: 11),   // 0000 1100 111     11      19
    (Code: $0B0; Lng: 11),   // 0000 1101 000     11      20
    (Code: $1B0; Lng: 11),   // 0000 1101 100     11      21
    (Code: $760; Lng: 11),   // 0000 0110 111     11      22
    (Code: $0A0; Lng: 11),   // 0000 0101 000     11      23
    (Code: $740; Lng: 11),   // 0000 0010 111     11      24
    (Code: $0C0; Lng: 11),   // 0000 0011 000     11      25
    (Code: $530; Lng: 12),   // 0000 1100 1010    12      26
    (Code: $D30; Lng: 12),   // 0000 1100 1011    12      27
    (Code: $330; Lng: 12),   // 0000 1100 1100    12      28
    (Code: $B30; Lng: 12),   // 0000 1100 1101    12      29
    (Code: $160; Lng: 12),   // 0000 0110 1000    12      30
    (Code: $960; Lng: 12),   // 0000 0110 1001    12      31
    (Code: $560; Lng: 12),   // 0000 0110 1010    12      32
    (Code: $D60; Lng: 12),   // 0000 0110 1011    12      33
    (Code: $4B0; Lng: 12),   // 0000 1101 0010    12      34
    (Code: $CB0; Lng: 12),   // 0000 1101 0011    12      35
    (Code: $2B0; Lng: 12),   // 0000 1101 0100    12      36
    (Code: $AB0; Lng: 12),   // 0000 1101 0101    12      37
    (Code: $6B0; Lng: 12),   // 0000 1101 0110    12      38
    (Code: $EB0; Lng: 12),   // 0000 1101 0111    12      39
    (Code: $360; Lng: 12),   // 0000 0110 1100    12      40
    (Code: $B60; Lng: 12),   // 0000 0110 1101    12      41
    (Code: $5B0; Lng: 12),   // 0000 1101 1010    12      42
    (Code: $DB0; Lng: 12),   // 0000 1101 1011    12      43
    (Code: $2A0; Lng: 12),   // 0000 0101 0100    12      44
    (Code: $AA0; Lng: 12),   // 0000 0101 0101    12      45
    (Code: $6A0; Lng: 12),   // 0000 0101 0110    12      46
    (Code: $EA0; Lng: 12),   // 0000 0101 0111    12      47
    (Code: $260; Lng: 12),   // 0000 0110 0100    12      48
    (Code: $A60; Lng: 12),   // 0000 0110 0101    12      49
    (Code: $4A0; Lng: 12),   // 0000 0101 0010    12      50
    (Code: $CA0; Lng: 12),   // 0000 0101 0011    12      51
    (Code: $240; Lng: 12),   // 0000 0010 0100    12      52
    (Code: $EC0; Lng: 12),   // 0000 0011 0111    12      53
    (Code: $1C0; Lng: 12),   // 0000 0011 1000    12      54
    (Code: $E40; Lng: 12),   // 0000 0010 0111    12      55
    (Code: $140; Lng: 12),   // 0000 0010 1000    12      56
    (Code: $1A0; Lng: 12),   // 0000 0101 1000    12      57
    (Code: $9A0; Lng: 12),   // 0000 0101 1001    12      58
    (Code: $D40; Lng: 12),   // 0000 0010 1011    12      59
    (Code: $340; Lng: 12),   // 0000 0010 1100    12      60
    (Code: $5A0; Lng: 12),   // 0000 0101 1010    12      61
    (Code: $660; Lng: 12),   // 0000 0110 0110    12      62
    (Code: $E60; Lng: 12));  // 0000 0110 0111    12      63

  // schwarz MakeUp-Code;  black MakeUp code
  cSFFMHBMU : TSFFMHMU = (
    (Code: $03C0; Lng: 10),   // 0000 0011 11       10      64
    (Code: $0130; Lng: 12),   // 0000 1100 1000     12      128
    (Code: $0930; Lng: 12),   // 0000 1100 1001     12      192
    (Code: $0DA0; Lng: 12),   // 0000 0101 1011     12      256
    (Code: $0CC0; Lng: 12),   // 0000 0011 0011     12      320
    (Code: $02C0; Lng: 12),   // 0000 0011 0100     12      384
    (Code: $0AC0; Lng: 12),   // 0000 0011 0101     12      448
    (Code: $06C0; Lng: 13),   // 0000 0011 0110 0   13      512
    (Code: $16C0; Lng: 13),   // 0000 0011 0110 1   13      576
    (Code: $0A40; Lng: 13),   // 0000 0010 0101 0   13      640
    (Code: $1A40; Lng: 13),   // 0000 0010 0101 1   13      704
    (Code: $0640; Lng: 13),   // 0000 0010 0110 0   13      768
    (Code: $1640; Lng: 13),   // 0000 0010 0110 1   13      832
    (Code: $09C0; Lng: 13),   // 0000 0011 1001 0   13      896
    (Code: $19C0; Lng: 13),   // 0000 0011 1001 1   13      960
    (Code: $05C0; Lng: 13),   // 0000 0011 1010 0   13      1024
    (Code: $15C0; Lng: 13),   // 0000 0011 1010 1   13      1088
    (Code: $0DC0; Lng: 13),   // 0000 0011 1011 0   13      1152
    (Code: $1DC0; Lng: 13),   // 0000 0011 1011 1   13      1216
    (Code: $0940; Lng: 13),   // 0000 0010 1001 0   13      1280
    (Code: $1940; Lng: 13),   // 0000 0010 1001 1   13      1344
    (Code: $0540; Lng: 13),   // 0000 0010 1010 0   13      1408
    (Code: $1540; Lng: 13),   // 0000 0010 1010 1   13      1472
    (Code: $0B40; Lng: 13),   // 0000 0010 1101 0   13      1536
    (Code: $1B40; Lng: 13),   // 0000 0010 1101 1   13      1600
    (Code: $04C0; Lng: 13),   // 0000 0011 0010 0   13      1664
    (Code: $14C0; Lng: 13));  // 0000 0011 0010 1   13      1728

  // Zustzliche MakeUp-Codes, nur in TIFF-Spezifikation enthalten
  // Extra MakeUp codes, only described in TIFF-Spezification
  cSFFMHEMU : TSFFMHEMU = (
    (Code: $0080; Lng: 11),   // 0000 0001 000     11      1792
    (Code: $0180; Lng: 11),   // 0000 0001 100     11      1856
    (Code: $0580; Lng: 11),   // 0000 0001 101     11      1920
    (Code: $0480; Lng: 12),   // 0000 0001 0010    12      1984
    (Code: $0C80; Lng: 12),   // 0000 0001 0011    12      2048
    (Code: $0280; Lng: 12),   // 0000 0001 0100    12      2112
    (Code: $0A80; Lng: 12),   // 0000 0001 0101    12      2176
    (Code: $0680; Lng: 12),   // 0000 0001 0110    12      2240
    (Code: $0E80; Lng: 12),   // 0000 0001 0111    12      2304
    (Code: $0380; Lng: 12),   // 0000 0001 1100    12      2368
    (Code: $0B80; Lng: 12),   // 0000 0001 1101    12      2432
    (Code: $0780; Lng: 12),   // 0000 0001 1110    12      2496
    (Code: $0F80; Lng: 12));  // 0000 0001 1111    12      2560

const
  cSFFMHTCs : array[Boolean] of PSFFMHTC = (@cSFFMHWTC, @cSFFMHBTC);
  cSFFMHMUs : array[Boolean] of PSFFMHMU = (@cSFFMHWMU, @cSFFMHBMU);

const
  cSFFPixelMask : array[1..16] of Word = (
     $0001, $0003, $0007, $000F,  $001F, $003F, $007F, $00FF,
     $01FF, $03FF, $07FF, $0FFF,  $1FFF, $3FFF, $7FFF, $FFFF);

function MHSearchTerminationCode(ABits:Word; var ABitCount, APixelCount:Integer;
                                 ABlackPixels:Boolean):Boolean;
var
  i : Integer;
  Table : PSFFMHTC;
begin
  Result:=False;
  Table:=cSFFMHTCs[ABlackPixels];
  for i:=0 to 63 do with Table^[i] do begin
    if (ABits and cSFFPixelMask[Lng])=Code then begin
      ABitCount:=Lng;
      APixelCount:=i;
      Result:=True;
      Exit;
    end;
  end;
end;

function MHSearchMakeUp(ABits:Word; var ABitCount, APixelCount:Integer;
                        ABlackPixels:Boolean):Boolean;
var
  i : Integer;
  Table : PSFFMHMU;
begin
  Result:=False;
  Table:=cSFFMHMUs[ABlackPixels];
  for i:=0 to 26 do with Table^[i] do begin
    if (ABits and cSFFPixelMask[Lng])=Code then begin
      ABitCount:=Lng;
      APixelCount:=(i+1)*64;
      Result:=True;
      Exit;
    end;
  end;
  // Zustzliche MakeUp-Tabelle fr Zeilen>1728
  // Extra MakeUp codes for lines > 1728 Pixel
  for i:=0 to 12 do with cSFFMHEMU[i] do begin
    if (ABits and cSFFPixelMask[Lng])=Code then begin
      ABitCount:=Lng;
      APixelCount:=1728+(i+1)*64;
      Result:=True;
      Exit;
    end;
  end;
end;

end.
