{$ALIGN ON, $BOOLEVAL OFF, $LONGSTRINGS ON, $IOCHECKS ON, $WRITEABLECONST OFF}
{$OVERFLOWCHECKS OFF, $RANGECHECKS OFF, $TYPEDADDRESS ON, $MINENUMSIZE 1}
unit SFF;

{ TSFFFile, TSFFGraphic Version 1.02

  Copyright:
     2000, Peter J. Haas, Barleben, Germany

    Bitte License.txt lesen;  Please read License.txt

  Kontakt/Contact:
    EMail:     PeterJHaas@t-online.de,
    HomePage:  http://home.t-online.de/home/PeterJHaas/delphi.htm

  Beschreibung/Description:
    sff.htm;  sff_en.htm

  History:
    Bitte Readme.txt lesen;  Please read Readme.txt

  Quellen/Source:
    Dokumentation 'CAPI 2.0 Part 1';  Documentation 'CAPI 2.0 Part 1'
}

interface
uses
  Windows, SysUtils, Classes, Graphics;

const
  SFF_File_ID = $66666653;   { SFFF }
  SFF_Page_ID = $FE;

type
  // Header des SFF-Dokumentes;  Header of the SFF document
  PSFFDocumentHeader = ^TSFFDocumentHeader;
  TSFFDocumentHeader = packed record
    sffSFF_ID                : DWord;
    sffVersion               : Byte;
    sffReserved1             : Byte;
    sffUserInformation       : Word;
    sffPageCount             : Word;
    sffOffsetFirstPageHeader : Word;
    sffOffsetLastPageHeader  : DWord;
    sffOffsetDocumentEnd     : DWord;
  end;

  // Header einer SFF-Seite;  Header of a SFF page
  PSFFPageHeader = ^TSFFPageHeader;
  TSFFPageHeader = packed record
    sffPageHeaderID          : Byte;
    sffPageHeaderLen         : Byte;
    sffResolutionVertical    : Byte;
    sffResolutionHorizontal  : Byte;
    sffCoding                : Byte;
    sffReserved2             : Byte;
    sffLineLength            : Word;
    sffPageLength            : Word;
    sffOffsetPreviousPage    : DWord;
    sffOffsetNextPage        : DWord;
  end;

  // Image fr TSFFGraphic, mehrere TSFFGraphic knnen sich ein Image teilen
  // Image for TSFFGraphic, multiple TSFFGraphic can share an image
  TSFFImage = class(TSharedImage)
  private
    // komprimierte Version, nur Pixeldaten; Nil, wenn nicht vorhanden
    // compressed Version, only pixel data; Nil, if not existent
    FData       : TCustomMemoryStream;
    // Anzahl der Zeilen;  Number of lines
    FLineCount  : Integer;
    function GetPageHeader:PSFFPageHeader;
    function GetPixelData:PByte;
    property PageHeader:PSFFPageHeader read GetPageHeader;
    property PixelData:PByte read GetPixelData;
  protected
    procedure FreeHandle; override;
  public
    destructor Destroy; override;
  end;

  TSFFGraphic = class(TGraphic)
  private
    FImage  : TSFFImage;   // komprimierte Daten;  compressed data
    FBitmap : TBitmap;     // unkomprimierte Version in einer Bitmap
                           // uncompressed Version in a bitmap
    // property-Methoden
    function GetResolutionVertical  :Integer;
    function GetResolutionHorizontal:Integer;
    function GetLineLength:Integer;
    function GetLineCount:Integer;
    procedure NewBitmap;
    procedure NewImage;
    // reine Daten streamen (ASize=-1, Gre unbekannt
    // stream pure data (ASize=-1, size unknown)
    procedure ReadStream (ASize:Integer; AStream: TStream);
    function GetBitmap:TBitmap;
  protected
    // TGraphic
    procedure AssignTo(Dest: TPersistent); override;
    procedure Draw(ACanvas: TCanvas; const Rect: TRect); override;
    function GetEmpty: Boolean; override;
    function GetPalette: HPALETTE; override;
    procedure SetPalette(Value: HPALETTE); override;
    function GetTransparent: Boolean; override;
    procedure SetTransparent(Value: Boolean); override;
    function GetWidth: Integer; override;
    procedure SetWidth(Value: Integer); override;
    function GetHeight: Integer; override;
    procedure SetHeight(Value: Integer); override;
    // Dekomprimierung;  Decompression
    procedure Decompress;
    property Bitmap:TBitmap read GetBitmap;
  public
    constructor Create; override;
    destructor Destroy; override;
    // TGraphic
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
    procedure ReadData(Stream: TStream); override;
    procedure WriteData(Stream: TStream); override;
    procedure LoadFromClipboardFormat(AFormat: Word; AData: THandle;
      APalette: HPALETTE); override;
    procedure SaveToClipboardFormat(var AFormat: Word; var AData: THandle;
      var APalette: HPALETTE); override;
    procedure Assign(Source: TPersistent); override;
    // Eigenschaften;  properties
    property LineCount:Integer read GetLineCount;
    // Resolution liefert 0, wenn nicht bekannt
    // Resolution returns 0, if unknown
    property ResolutionVertical  :Integer read GetResolutionVertical  ;
    property ResolutionHorizontal:Integer read GetResolutionHorizontal;
    property LineLength:Integer read GetLineLength;
  end;

  TSFFFile = class(TObject)
  private
    FList  : TList;     // Liste von Seiten;  List of pages
    FDocumentHeader : TSFFDocumentHeader; // Dokument-Header;  Document header
    FFilename : String;   // Dateiname;  file name
    // property-Methoden
    function GetPageCount:Integer;
    procedure SetFilename(const Value:String);
    function GetGraphics(AIndex:Integer):TSFFGraphic;
    // Header und Seitengren lesen;  read header and page sizes
    procedure ScanPages;
    procedure ClearPageList;
  protected
    function CreateStream:TStream; virtual;
  public
    constructor Create;
    destructor Destroy; override;
    // SFF public
    property PageCount:Integer read GetPageCount;
    property Filename:String read FFilename write SetFilename;
    property Graphics[AIndex:Integer]:TSFFGraphic read GetGraphics;
  end;

  ESFF = class(Exception);

implementation
uses
  MHTools, SFFConsts;


// ****************  TSFFImage  ****************

destructor TSFFImage.Destroy;
begin
  FData.Free;
  inherited Destroy;
end;

procedure TSFFImage.FreeHandle;
begin
end;

function TSFFImage.GetPageHeader:PSFFPageHeader;
begin
  if FData=Nil then raise ESFF.Create(SSFFEmptyImage);
  Result:=FData.Memory;
end;

function TSFFImage.GetPixelData:PByte;
begin
  if FData=Nil then raise ESFF.Create(SSFFEmptyImage);
  Result:=FData.Memory;
  if Result<>Nil then
    Inc(Result,PageHeader^.sffPageHeaderLen+2);
end;

// ****************  TSFFGraphic  ****************

constructor TSFFGraphic.Create;
begin
  inherited Create;
  NewImage;
  FBitmap:=TBitmap.Create;
end;

destructor TSFFGraphic.Destroy;
begin
  FBitmap.Free;
  FImage.Release;
  inherited Destroy;
end;

function TSFFGraphic.GetResolutionVertical:Integer;
const
  lpi : array[0..1] of Byte = (98,196);
begin
  if FImage.PageHeader^.sffResolutionVertical in [0..1] then
    Result:=lpi[FImage.PageHeader^.sffResolutionVertical]
  else
    Result:=0;
end;

function TSFFGraphic.GetResolutionHorizontal:Integer;
begin
  if FImage.PageHeader^.sffResolutionHorizontal=0 then
    Result:=203
  else
    Result:=0;
end;

function TSFFGraphic.GetLineLength:Integer;
begin
  Result:=FImage.PageHeader^.sffLineLength;
end;

function TSFFGraphic.GetLineCount:Integer;
begin
  Result:=FImage.FLineCount;
end;

procedure TSFFGraphic.Decompress;
var
  BufPtr  : PByte;        // aktueller Pointer;  actual Pointer
  DoubleLines : Boolean;  // doppelte Zeilen, um den Stretch-Effekt zu vemeiden
                          // double lines, to avoid the stretch effect

procedure DecompressLines;
var
  y : Integer;            // aktuelle Zeile;  actual line
  xLineLength : Integer;  // Zeilenbreite, gemerkt aus Header
                          // line width, remembered from the header

// einzelne Zeile dekomprimieren
procedure DecompressLine;

// ****************  Eingabe
var
  LineBufLen : Integer;   // verbleibene Bytes in Zeile;  remaining bytes in line
  LineBitPos : Integer;   // Nr. des aktuellen Bits;  No. of the actual bit

// BitStream fr eine Zeile vorbereiten
// prepare BitStream for a line
procedure LineReadBegin;
begin
  if BufPtr^=$00 then begin
    Inc(BufPtr);
    LineBufLen:=BufPtr^;
    Inc(BufPtr);
    LineBufLen:=LineBufLen or (BufPtr^ shl 8);
  end
  else begin
    LineBufLen:=BufPtr^;
  end;
  Inc(BufPtr);
  LineBitPos:=0;
end;

// 16 Bits aus BitStream lesen, ohne Zeiger weiterzusetzen
// read 16 bits from BitStream, without moving the pointer
function LineRead:Word;
var
  i : Integer;            // Anzahl Bit, die zu lesen sind
                          // number of bits to read
  k : Integer;            // Hilfsbyte;  help byte
  xBufPtr     : PByte;    // vorbergehender Zeiger;  temporary pointer
  xLineBitPos : Integer;  // vorbergehende Bit-Position
                          // temporary bit position
  xLineBufLen : Integer;  // vorbergehende verbleibende Byte-Anzahl
                          // temporary remaining byte count
begin
  Result:=0;
  if LineBufLen<=0 then Exit;  // keine Bits mehr vorhanden;  no bits left
  i:=0;
  xBufPtr:=BufPtr;
  xLineBitPos:=LineBitPos;
  xLineBufLen:=LineBufLen;
  while i<16 do begin          // 16 Bit lesen;  read 16 bit
    if xLineBufLen<=0 then begin
      k:=0;
    end
    else begin
      k:=xBufPtr^;
      if xLineBitPos>0 then k:=k shr xLineBitPos;
    end;
    // Daten bertragen;  transmit data
    if i>0 then k:=k shl i;
    Result:=Result or k;
    // Anzahl verringern;  decrease count
    Inc(i,8-xLineBitPos);
    // vorbergehende Zeiger weitersetzen;  move temporary pointer on
    xLineBitPos:=0;
    Inc(xBufPtr);
  end;
end;

// BitStream um ABitCount Bits weitersetzen
// move BitStream on for ABitCount bits
procedure LineReadSeek(ABitCount:Integer);
begin
  Inc(LineBitPos,ABitCount);   // Anzahl gelesener Bits erhhen
                               // increase number of read bits
  while LineBitPos>=8 do begin
    Dec(LineBitPos,8);
    Inc(BufPtr);
    Dec(LineBufLen);
  end;
  if (LineBufLen<0) or ((LineBufLen=0) and (LineBitPos>0)) then
    raise ESFF.Create(SSFFUnexpectedStreamEnd);
end;

procedure LineReadEnd;  // Puffer bis zum Ende lesen;  read buffer to the end
begin
  if LineBufLen>0 then
    Inc(BufPtr,LineBufLen);
end;

// ****************  Ausgabe;  Output
var
  // Bitmapausgabe;  Bitmap output
  BmpPointer1 : PByte;   // erste Zeile;  first line
  BmpPointer2 : PByte;   // Zeile verdoppeln;  double line
  BmpByte     : Byte;
  BmpPixelPos : Integer;

// Ausgabe vorbereiten;  prepare output
procedure LineWriteBegin(ALine:Integer);
begin
  BmpPointer1:=FBitmap.ScanLine[ALine*(Ord(DoubleLines)+1)];
  if DoubleLines then BmpPointer2:=FBitmap.ScanLine[ALine*2+1];
  BmpByte:=0;
  BmpPixelPos:=0;
end;

// Ein Pixel ausgeben;  output a pixel
procedure LineWrite(ABlackPixel:Boolean);
const
  cOutputPixelMask : array[0..7] of Byte = ($80,$40,$20,$10,$08,$04,$02,$01);
begin
  Assert(BmpPixelPos in [0..7]);
  if Not ABlackPixel then begin  // nur weie Pixel ausgeben (auf 1 setzen)
                                 // output only white pixel (set to 1)
    BmpByte:=BmpByte or cOutputPixelMask[BmpPixelPos];
  end;
  Inc(BmpPixelPos);
  if BmpPixelPos=8 then begin
    BmpPointer1^:=BmpByte;
    if DoubleLines then BmpPointer2^:=BmpByte;
    Inc(BmpPointer1);
    if DoubleLines then Inc(BmpPointer2);
    BmpByte:=0;
    BmpPixelPos:=0;
  end;
end;

// Ausgabe abschlieen (Puffer leeren);  end output (clear buffer)
procedure LineWriteEnd;
begin
  if BmpPixelPos>0 then begin
    BmpPointer1^:=BmpByte;
    if DoubleLines then
      BmpPointer2^:=BmpByte;
  end;
end;

var
  x : Integer;
  Bits        : Word;
  BitCount    : Integer;
  PixelCount  : Integer;
  MakeUpCount : Integer;
  BlackPixels : Boolean;
  i, j : Integer;
begin { DecompressLine }
  LineReadBegin;
  LineWriteBegin(y);
  x:=0;
  BlackPixels:=False;   // mit weien Pixeln anfangen;  begin with white pixels
  repeat
    Bits:=LineRead;     // nchste Bits lesen;  read next bits
    Assert(Bits<>0);
    if Not MHSearchTerminationCode(Bits,BitCount,PixelCount,BlackPixels) then begin
      // MakeUp testen;  test MakeUp
      if MHSearchMakeUp(Bits,BitCount,MakeUpCount,BlackPixels) then begin
        LineReadSeek(BitCount);
        Bits:=LineRead;
        if MHSearchTerminationCode(Bits,BitCount,PixelCount,BlackPixels) then begin
          Inc(PixelCount,MakeUpCount);
        end
        else begin
          PixelCount:=MakeUpCount;
          raise ESFF.Create(SSFFMakeUpWithoutTC);
        end;
      end
      else begin
        raise ESFF.Create(SSFFNoMakeUpOrTC);
      end;
    end;
    LineReadSeek(BitCount);
    j:=x+PixelCount-1;
    if j>=xLineLength then j:=xLineLength-1;
    for i:=x to j do LineWrite(BlackPixels);
    Inc(x,PixelCount);
    BlackPixels:=Not BlackPixels;  // Pixelfarbe umkehren;  reverse pixel color
  until x>=xLineLength;
  LineWriteEnd;
  LineReadEnd;
  Inc(y);
end;

// Leerzeilen;  empty lines
procedure ClearLines(ACount:Integer);
var
  i : Integer;
begin
  while ACount>0 do begin
    // Zeile fllen;  fill line
    i:=(xLineLength+7) div 8;
    FillChar(FBitmap.ScanLine[y*(Ord(DoubleLines)+1)]^  ,i,$FF);
    if DoubleLines then
      FillChar(FBitmap.ScanLine[y*2+1]^,i,$FF);
    // nchste Zeile;  next line
    Inc(y);
    Dec(ACount);
  end;
  Inc(BufPtr);
end;

// Kommentare berspringen;  skip comments
procedure JumpComments;
var
  i : Integer;
begin
  Inc(BufPtr);
  i:=BufPtr^;
  if i=0 then raise ESFF.Create(SSFFIllegalLineCoding);
  Inc(BufPtr,i+1);
end;

begin { DecompressLines }
  y:=0;
  xLineLength:=FImage.PageHeader^.sffLineLength;
  while True do begin
    case BufPtr^ of
      $00..$D8 : DecompressLine;           { einzelne Zeile mit Daten }
      $D9..$FD : ClearLines(BufPtr^-$D8);  { Leerzeilen }
      $FE : Break;                         { Seitenende }
      $FF : JumpComments;                  { Kommentare berspringen }
    end;
  end;
end;

begin { Decompress }
  if FImage.FLineCount=0 then Exit;   { keine Zeile => Dekomprimierung nicht notwendig }
  if FImage.FData.Size=0 then raise ESFF.Create(SSFFEmptyImage);
  BufPtr:=FImage.PixelData;
  try
    FBitmap.Width:=FImage.PageHeader^.sffLineLength;
    // Zeilen verdoppeln, um Stretch-Effekt zu kompensieren
    // double lines, to compensate stretch effect
    DoubleLines:=FImage.PageHeader^.sffResolutionVertical=0;
    FBitmap.Height:=FImage.FLineCount*(Ord(DoubleLines)+1);
    FBitmap.PixelFormat:=pf1Bit;
    DecompressLines;      // alle Zeilen dekomprimieren;  decompress all lines
  except
    FBitmap.Assign(Nil);  // bei Fehler die Graphic lschen
                          // on error clear graphic
    raise;
  end;
end;

procedure TSFFGraphic.ReadData(Stream: TStream);
var
  Size : Longint;
begin
  Stream.Read(Size,SizeOf(Size));
  ReadStream(Size,Stream);
end;

procedure TSFFGraphic.WriteData(Stream: TStream);
var
  Size: Longint;
begin
  if (FImage<>Nil) and (FImage.FData<>Nil) then
    Size:=FImage.FData.Size
  else
    Size:=0;
  Stream.Write(Size,SizeOf(Size));
  if Size>0 then Stream.Write(FImage.FData.Memory^,Size);
end;

procedure TSFFGraphic.ReadStream(ASize:Integer; AStream:TStream);
var
  // Lesepuffer;  ReadBuffer
  Buffer : Pointer;           // Pointer auf Lesepuffer
                              // Pointer on read buffer
  BufEnd : Integer;           // Anzahl Zeichen im Lesepuffer
                              // Number of chars in the read buffer
  BufPtr : PByte;
  StreamPosition : Integer;   // Position im Stream bei Beginn des Blockes
                              // Position in the stream at the beginning of the block
  PageOffset : Integer;       // Beginn der Imagedaten im Lesepuffer
                              // Beginning of the image data in the read buffer
  xLineCount : Integer;       // Anzahl Zeilen, gemerkt
                              // number of lines, remebered   

// Imagedaten -> Image kopieren;  Image data -> copy image
procedure WriteImageData(ABufPos:Integer);
begin
  if (PageOffset>=0) and (PageOffset<ABufPos) then begin
    if FImage.FData=Nil then
      FImage.FData:=TMemoryStream.Create;
    FImage.FData.WriteBuffer(Pointer(LongInt(Buffer)+PageOffset)^,
                             ABufPos-PageOffset);
    PageOffset:=0;
  end;
end;

// nchsten Block aus Stream lesen;  read next block from the stream
procedure ReadNextBlock;
begin
  WriteImageData(BufEnd);   // Pixeldaten kopieren;  copy pixel data
  StreamPosition:=AStream.Position;
  BufEnd:=AStream.Read(Buffer^,ASize);
  BufPtr:=Buffer;
  if BufEnd=0 then raise ESFF.Create(SSFFInvalidFile); // vorzeitiges Ende
                                                       // premature end
end;

procedure iReadBuffer(ABuffer:PByte; ACount:Integer);
var
  i : Integer;
begin
  if ACount<=0 then Exit;
  // Anzahl restlicher Zeichen im Puffer berechnen
  // calculate number of remaining chars in the buffer
  i:=BufEnd-(LongInt(BufPtr)-LongInt(Buffer));
  while ACount>0 do begin
    if i=0 then begin  // Block zu Ende;  block end
      ReadNextBlock;
      i:=BufEnd;
    end;
    if i>ACount then i:=ACount; // Anzahl aus akt. Block zu lesender Zeichen
                                // number of chars to read from the actual block
    if ABuffer<>Nil then Move(BufPtr^,ABuffer^,i);
    Inc(BufPtr,i);
    if ABuffer<>Nil then Inc(ABuffer,i);
    Dec(ACount,i);
    i:=0;
  end;
end;

procedure ReadBuffer(var ABuffer; ACount:Integer);
begin
  iReadBuffer(@ABuffer,ACount);
end;

procedure ReadWithoutCopy(ACount:Integer);
begin
  iReadBuffer(Nil,ACount);
end;

// Kopieren der Pixeldaten beenden
// stop copying the pixel data
procedure EndOfCopy;
var
  i : Integer;
begin
  i:=LongInt(BufPtr)-LongInt(Buffer);
  WriteImageData(i);
  AStream.Position:=StreamPosition+i-1; // Stream auf Anfang der nchsten Seite
                                        // set stream onto the beginning of the next page
end; { EndOfCopy }

procedure ScanPage;
var
  i : Integer;
  b, b2 : Byte;
  w : Word;
begin
  FImage.FLineCount:=0;
  xLineCount:=0;
  // Kopierstart festlegen;  Set copy start
  PageOffset:=LongInt(BufPtr)-LongInt(Buffer);
  // Kennbyte fr Seite lesen;  read code byte for page
  ReadBuffer(b,1);
  // Lnge des Headers lesen;  read length of header
  ReadBuffer(b2,1);
  if (b=SFF_Page_ID) and (b2=0) then Exit;        // Document end
  if (b<>SFF_Page_ID) or (b2<SizeOf(TSFFPageHeader)-2) then
    raise ESFF.Create(SSFFInvalidFile);
  // Header berspringen;  skip header
  ReadWithoutCopy(b2);

  // Zeilen scannen;  scan lines
  i:=0;
  while True do begin
    ReadBuffer(b,SizeOf(b));  // Einleitungsbyte lesen;  read initiation byte
    case b of
      $00 : begin                 // Escape, Word mit Lnge folgt
                                  // Escape, word with length is following
        ReadBuffer(w,SizeOf(w));
        i:=w;
        Inc(xLineCount);
      end;
      $01..$D8 : begin   // Lnge;  Length
        i:=b;
        Inc(xLineCount);
      end;
      $D9..$FD : begin   // Leerzeilen;  empty lines
        i:=0;
        Inc(xLineCount,b-$D8);
      end;
      $FE : Break;       // Ende der Seite;  end of page
      $FF : begin        // Zusatzdaten;  additional data
        ReadBuffer(b,SizeOf(b));
        i:=b;
        if i=0 then raise ESFF.Create(SSFFIllegalLineCoding);
      end;
    end;
    ReadWithoutCopy(i);  // i Bytes weitersetzen;  move on for i bytes
  end;
  EndOfCopy;
  FImage.FLineCount:=xLineCount;
end; { ScanPage }

const
  cBufferSize = 32768;
begin  { ReadStream }
  if ASize=0 then Exit;
  if ASize<0 then ASize:=cBufferSize;
  GetMem(Buffer,ASize);
  try
    NewImage;
    PageOffset:=-1;            // noch keine Pixeldaten zu schreiben
                               // don't write pixel data, yet
    ReadNextBlock;  // ersten Block lesen;  read first block
    // Fileheader auswerten, wenn vorhanden
    // evaluate file header, if existent
    if BufPtr^<>$FE then begin   // zuerst DocumentHeader;  first document header
      with PSFFDocumentHeader(BufPtr)^ do begin
        if (BufEnd<SizeOf(TSFFDocumentHeader)) or
           (sffSFF_ID<>SFF_File_ID) or
           (sffVersion<>1) or
           (sffUserInformation<>0) then raise ESFF.Create(SSFFInvalidFile);
        ReadWithoutCopy(sffOffsetFirstPageHeader);  // Lesepuffer weitersetzen
                                                    // move on read buffer
      end;
    end;
    ScanPage;  // Seite ab Header scannen;  scan page beginning with the header
  finally
    FreeMem(Buffer,ASize);
  end;
  Changed(Self);
end; { ReadStream }

procedure TSFFGraphic.LoadFromStream(Stream: TStream);
begin
  ReadStream(-1,Stream);  // unbekannte Lnge;  unknown length
end;

procedure TSFFGraphic.SaveToStream(Stream: TStream);
begin
  raise ESFF.Create(SSFFNoSave);
end;

procedure TSFFGraphic.LoadFromClipboardFormat(AFormat: Word; AData: THandle;
      APalette: HPALETTE);
begin
  raise ESFF.Create(SSFFToClipboard);
end;

procedure TSFFGraphic.SaveToClipboardFormat(var AFormat: Word; var AData: THandle;
      var APalette: HPALETTE);
begin
  raise ESFF.Create(SSFFToClipboard);
end;

procedure TSFFGraphic.Draw(ACanvas: TCanvas; const Rect: TRect);
begin
  if FImage=Nil then Exit;
  ACanvas.StretchDraw(Rect,Bitmap);
end;

function TSFFGraphic.GetEmpty: Boolean;
begin
  Result:=(FImage=Nil) and FBitmap.Empty;
end;

function TSFFGraphic.GetWidth: Integer;
begin
  Result:=0;
  if FImage<>Nil then begin
    Result:=Bitmap.Width;
  end;
end;

function TSFFGraphic.GetHeight: Integer;
begin
  Result:=0;
  if FImage<>Nil then begin
    Result:=Bitmap.Height;
  end;
end;

procedure TSFFGraphic.SetWidth(Value: Integer);
begin
  raise ESFF.Create(SSFFChangeSize);
end;

procedure TSFFGraphic.SetHeight(Value: Integer);
begin
  raise ESFF.Create(SSFFChangeSize);
end;

// Palette
function TSFFGraphic.GetPalette: HPALETTE;
begin
  Result:=FBitmap.Palette;
end;

procedure TSFFGraphic.SetPalette(Value: HPALETTE);
begin
  FBitmap.Palette:=Value;
end;

// Transparenz;  Transparency
function TSFFGraphic.GetTransparent: Boolean;
begin
  Result:=FBitmap.Transparent;
end;

procedure TSFFGraphic.SetTransparent(Value: Boolean);
begin
  FBitmap.Transparent:=Value;
end;

function TSFFGraphic.GetBitmap:TBitmap;
begin
  if FBitmap.Empty then Decompress;
  Result:=FBitmap;
end;

procedure TSFFGraphic.NewBitmap;
begin
  FBitmap.Assign(Nil);
end;

procedure TSFFGraphic.NewImage;
begin
  if FImage<>Nil then FImage.Release;
  FImage:=TSFFImage.Create;
  FImage.Reference;
end;

procedure TSFFGraphic.Assign(Source: TPersistent);
begin
  if Source is TSFFGraphic then begin
    FImage.Release;
    FImage:=TSFFGraphic(Source).FImage;
    FImage.Reference;
    if TSFFGraphic(Source).FBitmap<>Nil then begin
      NewBitmap;
      FBitmap.Assign(TSFFGraphic(Source).FBitmap);
    end;
  end
  else begin
    inherited Assign(Source);
  end;
end;

procedure TSFFGraphic.AssignTo(Dest: TPersistent);
begin
  if Dest is TBitmap then begin
    Dest.Assign(Bitmap);
  end
  else begin
    inherited AssignTo(Dest);
  end;
end;

// ****************  TSFFFile  ****************

constructor TSFFFile.Create;
begin
  inherited Create;
  FList:=TList.Create;
end;

destructor TSFFFile.Destroy;
begin
  ClearPageList;
  FList.Free;
  inherited Destroy;
end;

// Lschen der Liste und Freigeben der alten Dateien
// clear list and remove old files
procedure TSFFFile.ClearPageList;
var
  i : Integer;
begin
  with FList do begin
    for i:=0 to Count-1 do begin
      TObject(Items[i]).Free;
    end;
    Clear;
  end;
end;

function TSFFFile.GetPageCount:Integer;
begin
  Result:=FList.Count;
end;

function TSFFFile.CreateStream:TStream;
begin
  Result:=TFileStream.Create(Filename,fmOpenRead or fmShareDenyWrite);
end;

// Alle Seiten ermitteln und in Liste einfgen
// get all pages and insert in list
procedure TSFFFile.ScanPages;
var
  xStream : TStream;
  b       : Byte;
  Graphic : TSFFGraphic;
begin
  xStream:=CreateStream;
  if Assigned(xStream) then try
    xStream.ReadBuffer(FDocumentHeader,SizeOf(TSFFDocumentHeader));
    if FDocumentHeader.sffSFF_ID<>SFF_File_ID then
      raise Exception.Create(SSFFInvalidFile);
    xStream.Seek(FDocumentHeader.sffOffsetFirstPageHeader,0);
    repeat
      // Page-Kennung prfen;  check page code
      xStream.ReadBuffer(b,SizeOf(b));
      if b<>SFF_Page_ID then raise Exception.Create(SSFFInvalidFile);
      // Headerlnge bzw. Ende prfen
      // check headerlength respectively end
      xStream.ReadBuffer(b,SizeOf(b));
      if b<>0 then begin     // neue Seite;  new page
        xStream.Seek(-2,1);  // zurck auf Seitenanfang;  back to page origin
        Graphic:=TSFFGraphic.Create;
        FList.Add(Graphic);
        Graphic.LoadFromStream(xStream);
      end;
    until b=0;
  finally
    xStream.Free;
  end;
end;

procedure TSFFFile.SetFilename(const Value:String);
begin
  if FFilename<>Value then begin
    ClearPageList;
    FFilename:=Value;
    if (Length(FFilename)<>0) and FileExists(FFilename) then ScanPages;
  end;
end;

function TSFFFile.GetGraphics(AIndex:Integer):TSFFGraphic;
begin
  Result:=Nil;
  if (AIndex>=0) and (AIndex<FList.Count) then begin
    Result:=FList.Items[AIndex];
  end;
end;

initialization
  TPicture.RegisterFileFormat('sff', SSFFSFF, TSFFGraphic);
finalization
  TPicture.UnregisterGraphicClass(TSFFGraphic);
end.
